# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import random

import pandas as pd

from probatus.utils import NotInstalledError

from ..utils import assure_numpy_array

try:
    from scipy import stats
except ModuleNotFoundError:
    stats = NotInstalledError("scipy", "extras")


def sw(d1, d2, verbose=False):
    """
    Calculates the Shapiro-Wilk test statistic on 2 distributions.

    This examines whether deviation from normality of two distributions are significantly different.

    References:

    - [Wikipedia article about the Shapiro-Wilk test](https://en.wikipedia.org/wiki/Shapiro%E2%80%93Wilk_test)
    - [SciPy documentation](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.shapiro.html)


    Args:
        d1 (np.ndarray or pandas.Series): First sample.

        d2 (np.ndarray or pandas.Series): Second sample.

        verbose (bool): If True, useful interpretation info is printed to stdout.

    Returns:
        float: Shapiro-Wilk test statistic
        float: p-value of rejecting the null hypothesis (that the two distributions are identical)
    """
    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    if len(d1) > 5000:
        d1 = pd.Series(random.choices(d1, k=5000))
    if len(d2) > 5000:
        d2 = pd.Series(random.choices(d2, k=5000))

    delta = stats.shapiro(d1)[0] - stats.shapiro(d2)[0]

    d1 = pd.Series(d1)
    d2 = pd.Series(d2)

    MOT = pd.concat([d1, d2])
    n1 = d1.shape[0]
    n2 = d2.shape[0]

    def ran_delta(n1, n2):
        take_ran = lambda n: random.sample(range(MOT.shape[0]), n)
        ran_1 = MOT.iloc[take_ran(n1),]
        ran_2 = MOT.iloc[take_ran(n2),]
        delta_ran = stats.shapiro(ran_1)[0] - stats.shapiro(ran_2)[0]
        return delta_ran

    collect = [ran_delta(n1, n2) for a in range(100)]
    collect = pd.Series(list(collect))
    delta_p_value = 1 - stats.percentileofscore(collect, delta) / 100

    quants = [0.025, 0.975]
    sig_vals = list(collect.quantile(quants))

    if verbose:
        if delta < sig_vals[0] or delta > sig_vals[1]:
            print("\nShapiro_Difference | Null hypothesis : <delta is not different from zero> REJECTED.")
            print("\nDelta is outside 95% CI -> Distributions very different.")
        else:
            print("\nShapiro_Difference | Null hypothesis : <delta is not different from zero> NOT REJECTED.")
            print("\nDelta is inside 95% CI -> Distributions are not different.")

    return delta, delta_p_value



# This approach is adapted from, and explained in: https://calmcode.io/docs/epic.html

import os

import pytest

# Turn off interactive mode in plots

plots_disable = "import matplotlib \n" "import matplotlib.pyplot as plt \n" "plt.ioff() \n" "matplotlib.use('Agg') \n"
NOTEBOOKS_PATH = "./docs/tutorials/"

NOTEBOOKS_TO_TEST_LGBM = [
    "./docs/tutorials/nb_shap_feature_elimination",
    "./docs/tutorials/nb_imputation_comparision",
    "./docs/discussion/nb_rfecv_vs_shaprfecv.ipynb",
]

NOTEBOOKS_TO_TEST = [
    "./docs/tutorials/nb_binning",
    "./docs/tutorials/nb_custom_scoring",
    "./docs/tutorials/nb_distribution_statistics",
    "./docs/tutorials/nb_metric_volatility",
    "./docs/tutorials/nb_sample_similarity",
    "./docs/tutorials/nb_shap_model_interpreter" "./docs/howto/reproducibility",
]


def execute_notebook_test(notebook_name):
    """
    Execute a notebook.
    """
    notebook_path = notebook_name + ".ipynb"

    code_to_execute = os.popen(
        f"jupyter nbconvert --to script --execute --stdout --PythonExporter.exclude_markdown=True {notebook_path}"
    ).read()
    _ = os.popen(f"python3 -c {plots_disable + code_to_execute}").read()


@pytest.mark.parametrize("notebook_name", NOTEBOOKS_TO_TEST_LGBM)
@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
@pytest.mark.skip(reason="GitHub pipelines are failing on these tests. Could not find a solution for now.")
def test_jupyter_notebook_lgbm(notebook_name):
    """
    Test a notebook.
    """
    execute_notebook_test(notebook_name)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from .exceptions import NotFittedError, DimensionalityError, UnsupportedModelError, NotInstalledError
from .scoring import Scorer, get_scorers, get_single_scorer
from .arrayfuncs import (
    assure_numpy_array,
    assure_pandas_df,
    check_1d,
    check_numeric_dtypes,
    assure_pandas_series,
    preprocess_data,
    preprocess_labels,
)
from .shap_helpers import shap_calc, shap_to_df, calculate_shap_importance
from .warnings import ApproximationWarning
from ._utils import (
    class_name_from_object,
    assure_list_of_strings,
    assure_list_values_allowed,
)
from .plots import plot_distributions_of_feature
from .interface import BaseFitComputeClass, BaseFitComputePlotClass

__all__ = [
    "NotFittedError",
    "DimensionalityError",
    "UnsupportedModelError",
    "NotInstalledError",
    "Scorer",
    "assure_numpy_array",
    "assure_pandas_df",
    "check_1d",
    "ApproximationWarning",
    "class_name_from_object",
    "get_scorers",
    "assure_list_of_strings",
    "assure_list_values_allowed",
    "check_numeric_dtypes",
    "plot_distributions_of_feature",
    "shap_calc",
    "shap_to_df",
    "calculate_shap_importance",
    "assure_pandas_series",
    "preprocess_data",
    "preprocess_labels",
    "BaseFitComputeClass",
    "BaseFitComputePlotClass",
    "get_single_scorer",
]



import numpy as np


def generate_MCAR(df, missing):
    """
    Generate missing values completely at random for dataframe df.

    Args:
        df: input dataframe where some values will be masked
        missing: (float or dict)
            - float ( must be a fraction between 0 and 1 - both inclusive), then it will apply this
            fraction of missing values on the whole dataset.
            - dict:
                - keys: column names to mask values
                - values: fraction of missing values for this column

    Returns:
        pd.DataFrame: same as the input dataframe, but with some values masked based on the missing variable

    Examples:

        # Apply 20% missing values over all the columns
        miss_rand = generate_MCAR(data, missing=0.2)

        # Use the dictionary
        missing_vals = {"PAY_0":0.3,"PAY_5": 0.5}
        miss_rand = generate_MCAR(data, missing=missing_vals)

    """

    df = df.copy()

    if type(missing) == float and missing <= 1 and missing >= 0:
        df = df.mask(np.random.random(df.shape) < missing)
    elif type(missing) == dict:
        for k, v in missing.items():
            df[k] = df[k].mask(np.random.random(df.shape[0]) < v)

    else:
        raise ValueError("missing must be float within range [0.1] or dict")

    return df



from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from probatus.interpret.inspector import BaseInspector, InspectorShap, return_confusion_metric
from probatus.utils import NotFittedError, UnsupportedModelError
from tests.mocks import MockClusterer, MockModel

test_sensitivity = 0.0000000001


@pytest.mark.skip(reason="Not currently implemented")
def test_after_implementation_completed():
    """
    Test.
    """

    @pytest.fixture(scope="function")
    def global_clusters():
        return pd.Series([1, 2, 3, 4, 1, 2, 3, 4], name="cluster_id")

    @pytest.fixture(scope="function")
    def global_clusters_eval_set():
        return [pd.Series([1, 2, 3], name="cluster_id"), pd.Series([1, 2, 3], name="cluster_id")]

    @pytest.fixture(scope="function")
    def global_y():
        return pd.Series([0, 1, 1, 0, 0, 0, 1, 0])

    @pytest.fixture(scope="function")
    def global_X():
        return pd.DataFrame([[0], [1], [1], [0], [0], [0], [1], [0]])

    @pytest.fixture(scope="function")
    def global_confusion_metric():
        return pd.Series([0.1, 0.8, 0.3, 0.1, 0.1, 0.3, 0.3, 0.1])

    @pytest.fixture(scope="function")
    def global_summary_df(columns_summary_df):
        return pd.DataFrame(
            [
                [1, 0, 0.1, 0.1],
                [2, 1, 0.2, 0.8],
                [3, 1, 0.7, 0.3],
                [4, 0, 0.1, 0.1],
                [1, 0, 0.1, 0.1],
                [2, 0, 0.3, 0.3],
                [3, 1, 0.7, 0.3],
                [4, 0, 0.1, 0.1],
            ],
            columns=columns_summary_df,
        )

    @pytest.fixture(scope="function")
    def global_X_shap():
        return pd.DataFrame(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
                [2, 0, 0, 0],
                [0, 2, 0, 0],
                [0, 0, 2, 0],
                [0, 0, 0, 2],
            ],
            columns=["shap1", "shap2", "shap3", "shap4"],
        )

    @pytest.fixture(scope="function")
    def columns_aggregate_summary_df():
        return [
            "cluster_id",
            "total_label_1",
            "total_entries",
            "label_1_rate",
            "average_confusion",
            "average_pred_proba",
        ]

    @pytest.fixture(scope="function")
    def columns_summary_df():
        return ["cluster_id", "target", "pred_proba", "confusion"]

    @pytest.fixture(scope="function")
    def global_aggregate_summary_df(columns_aggregate_summary_df):
        return pd.DataFrame(
            [[1, 0, 2, 0, 0.1, 0.1], [2, 1, 2, 0.5, 0.55, 0.25], [3, 2, 2, 1, 0.3, 0.7], [4, 0, 2, 0, 0.1, 0.1]],
            columns=columns_aggregate_summary_df,
        )

    @pytest.fixture(scope="function")
    def global_aggregate_summary_dfs_eval_set(columns_aggregate_summary_df):
        return [
            pd.DataFrame(
                [[1, 0, 1, 0, 0.1, 0.1], [2, 0, 1, 0, 0.2, 0.2], [3, 0, 1, 0, 0.3, 0.3]],
                columns=columns_aggregate_summary_df,
            ),
            pd.DataFrame(
                [[1, 1, 1, 1, 0.4, 0.6], [2, 1, 1, 1, 0.5, 0.5], [3, 1, 1, 1, 0.6, 0.4]],
                columns=columns_aggregate_summary_df,
            ),
        ]

    @pytest.fixture(scope="function")
    def global_summary_dfs():
        return [
            pd.DataFrame([[1, 2, 3], [2, 3, 4], [3, 4, 5]], columns=["cluster_id", "column_a", "column_b"]),
            pd.DataFrame([[1, 2, 1], [2, 3, 2], [3, 4, 3]], columns=["cluster_id", "column_a", "column_b"]),
        ]

    @pytest.fixture(scope="function")
    def global_X_shaps():
        return [
            pd.DataFrame([[0, 3, 0], [3, 0, 0], [0, 0, 3]], columns=["shap_1", "shap_2", "shap_3"]),
            pd.DataFrame([[0, 2, 0], [2, 0, 0], [0, 0, 2]], columns=["shap_1", "shap_2", "shap_3"]),
        ]

    @pytest.fixture(scope="function")
    def global_ys():
        return [pd.Series([0, 0, 0]), pd.Series([1, 1, 1])]

    @pytest.fixture(scope="function")
    def global_Xs():
        return [pd.DataFrame([[0], [1], [1]]), pd.DataFrame([[0], [1], [1]])]

    @pytest.fixture(scope="function")
    def global_predicted_probas():
        return [pd.Series([0.1, 0.2, 0.3]), pd.Series([0.4, 0.5, 0.6])]

    @pytest.fixture(scope="function")
    def global_predicted_proba():
        return pd.Series([0.1, 0.2, 0.7, 0.1, 0.1, 0.3, 0.7, 0.1], name="pred_proba")

    @pytest.fixture(scope="function")
    def global_small_df():
        return pd.DataFrame([[1, 2, 3, 4], [1, 2, 3, 4]])

    @pytest.fixture(scope="function")
    def global_small_df_flat():
        return pd.Series([1, 2, 3, 4])

    @pytest.fixture(scope="function")
    def global_mock_aggregate_summary_dfs():
        return [
            pd.DataFrame([[1, 3], [2, 3]], columns=["cluster_id", "column_a"]),
            pd.DataFrame([[1, 2], [2, 3]], columns=["cluster_id", "column_a"]),
        ]

    @pytest.fixture(scope="function")
    def global_mock_summary_df():
        return pd.DataFrame([[1, 2], [2, 3]], columns=["cluster_id", "column_a"])

    def test_return_confusion_metric__array():
        y_true = np.array([0, 0, 0, 1, 1, 1], dtype=float)
        y_score = np.array([0.1, 0.2, 0.3, 0.7, 0.8, 0.9], dtype=float)

        expected_output_not_normalized = np.array([0.1, 0.2, 0.3, 0.3, 0.2, 0.1], dtype=float)
        expected_output_normalized = np.array(
            [0.11111111, 0.22222222, 0.33333333, 0.22222222, 0.11111111, 0.0], dtype=float
        )
        assert (
            expected_output_normalized - return_confusion_metric(y_true, y_score, normalize=True < test_sensitivity)
        ).all()
        assert (
            expected_output_not_normalized - return_confusion_metric(y_true, y_score, normalize=False)
            < test_sensitivity
        ).all()

    def test_return_confusion_metric__series():
        # The method also needs to work with series, since it is called with series by create summary df
        y_true = pd.Series([0, 0, 0, 1, 1, 1])
        y_score = pd.Series([0.1, 0.2, 0.3, 0.7, 0.8, 0.9])

        expected_output_not_normalized = pd.Series([0.1, 0.2, 0.3, 0.3, 0.2, 0.1], dtype=float)
        expected_output_normalized = pd.Series(
            [0.11111111, 0.22222222, 0.33333333, 0.22222222, 0.11111111, 0.0], dtype=float
        )
        assert (
            expected_output_normalized - return_confusion_metric(y_true, y_score, normalize=True < test_sensitivity)
        ).all()
        assert (
            expected_output_not_normalized - return_confusion_metric(y_true, y_score, normalize=False)
            < test_sensitivity
        ).all()

    @patch.object(MockClusterer, "fit")
    def test_fit_clusters__base_inspector(mock_clusterer, global_small_df):
        # Base Inspector case algotype is kmeans
        inspector = BaseInspector(algotype="kmeans")
        inspector.clusterer = mock_clusterer

        X = global_small_df

        inspector.fit_clusters(X)

        # Check if has been called with correct argument
        mock_clusterer.fit.assert_called_once()
        pd.testing.assert_frame_equal(mock_clusterer.fit.call_args[0][0], X)
        # Check if it has not been modified
        pd.testing.assert_frame_equal(X, global_small_df)
        # Check if fitted flag has been changed correctly
        assert inspector.fitted is True

    @patch.object(MockClusterer, "fit")
    def test_fit_clusters__inspector_shap(mock_clusterer, global_small_df):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        inspector.clusterer = mock_clusterer

        X = global_small_df

        inspector.fit_clusters(X)

        # Check if has been called with correct argument
        mock_clusterer.fit.assert_called_once()
        pd.testing.assert_frame_equal(mock_clusterer.fit.call_args[0][0], X)
        # Check if it has not been modified
        pd.testing.assert_frame_equal(X, global_small_df)
        # Check if fitted flag has been changed correctly
        assert inspector.fitted is True

    @patch.object(MockClusterer, "fit")
    def test_fit_clusters__inspector_shap_proba(mock_clusterer, global_small_df):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=True)
        inspector.clusterer = mock_clusterer
        inspector.predicted_proba = True

        X = global_small_df

        # Check if not fitted exception is raised
        inspector.fit_clusters(X)

        # Check if column with probabilities has been added to the fitted X
        assert "probs" in mock_clusterer.fit.call_args[0][0].columns

        # Check if has been called
        mock_clusterer.fit.assert_called_once()

        # Check if X has not been modified
        pd.testing.assert_frame_equal(X, global_small_df)
        assert inspector.fitted is True

    @patch.object(MockClusterer, "predict")
    def test_predict_clusters__base_inspector(mock_clusterer, global_small_df):
        mock_clusterer.predict.return_value = [1, 0]

        inspector = BaseInspector(algotype="kmeans")
        inspector.clusterer = mock_clusterer
        inspector.fitted = True

        X = global_small_df

        # Check if the prediction is correct according to the Mock clusterer
        assert inspector.predict_clusters(X) == [1, 0]

        # Check if the clusterer was called with correct input
        mock_clusterer.predict.assert_called_once()
        pd.testing.assert_frame_equal(mock_clusterer.predict.call_args[0][0], X)

        # Check if the X has not been modified
        pd.testing.assert_frame_equal(X, global_small_df)

    @patch.object(MockClusterer, "predict")
    def test_predict_clusters__inspector_shap(mock_clusterer, global_small_df):
        mock_clusterer.predict.return_value = [1, 0]

        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        inspector.clusterer = mock_clusterer
        inspector.fitted = True

        X = global_small_df

        # Check if the output is correct, as should be according to MockClusterer
        assert inspector.predict_clusters(X) == [1, 0]
        # Check if the df has not been modified by the prediction
        pd.testing.assert_frame_equal(X, global_small_df)

    @patch.object(MockClusterer, "predict")
    def test_predict_clusters__not_fitted(mock_clusterer, global_small_df):
        mock_clusterer.predict.return_value = [1, 0]

        # InspectorShap not fitted
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=True)
        inspector.clusterer = mock_clusterer
        inspector.predicted_proba = True

        X = global_small_df

        # Check if not fitted exception is raised
        with pytest.raises(NotFittedError):
            inspector.predict_clusters(X)
        # Check if X3 has not been modified
        pd.testing.assert_frame_equal(X, global_small_df)

    def test_assert_is_dataframe(global_small_df):
        X_df = global_small_df
        X_list = X_df.values.tolist()
        X_array = np.asarray(X_list)
        X_array_flat = np.asarray(X_list[0])

        pd.testing.assert_frame_equal(X_df, BaseInspector.assert_is_dataframe(X_df))
        pd.testing.assert_frame_equal(X_df, BaseInspector.assert_is_dataframe(X_array))
        with pytest.raises(NotImplementedError):
            BaseInspector.assert_is_dataframe(X_list)
        with pytest.raises(NotImplementedError):
            BaseInspector.assert_is_dataframe(X_array_flat)

    def test_assert_is_series(global_small_df, global_small_df_flat):
        X_df = global_small_df
        X_df_flat = global_small_df_flat
        X_list = X_df.values.tolist()
        X_list_flat = X_df_flat.values.tolist()

        X_series = pd.Series(X_list_flat)
        X_array = np.asarray(X_list)
        X_array_flat = np.asarray(X_list_flat)
        index = [0, 1, 2, 3]

        pd.testing.assert_series_equal(X_series, BaseInspector.assert_is_series(X_series))
        pd.testing.assert_series_equal(X_series, BaseInspector.assert_is_series(X_df_flat))
        pd.testing.assert_series_equal(X_series, BaseInspector.assert_is_series(X_array_flat, index=index))

        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_list)
        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_list_flat)
        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_df)
        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_array)
        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_array, index=[0, 1])
        with pytest.raises(TypeError):
            BaseInspector.assert_is_series(X_array_flat)

    def test_get_cluster_mask(global_summary_df):
        df = global_summary_df
        cluster_id_1 = 1
        cluster_id_2 = [1, 4]

        expected_indexes_1 = [0, 4]
        expected_indexes_2 = [0, 3, 4, 7]

        pd.testing.assert_frame_equal(df.iloc[expected_indexes_1], df[InspectorShap.get_cluster_mask(df, cluster_id_1)])
        pd.testing.assert_frame_equal(df.iloc[expected_indexes_2], df[InspectorShap.get_cluster_mask(df, cluster_id_2)])

    @patch("probatus.interpret.inspector.return_confusion_metric")
    def test_create_summary_df(
        mocked_method, global_clusters, global_y, global_predicted_proba, global_confusion_metric, global_summary_df
    ):
        cluster_series = global_clusters
        y_series = global_y
        probas = global_predicted_proba

        mocked_method.return_value = global_confusion_metric
        expected_output = global_summary_df

        output = InspectorShap.create_summary_df(cluster_series, y_series, probas, normalize=False)

        # Check if method is called with correct input
        mocked_method.assert_called_once()
        pd.testing.assert_series_equal(mocked_method.call_args[0][0], y_series)
        pd.testing.assert_series_equal(mocked_method.call_args[0][1], probas)
        assert mocked_method.call_args_list[0][1]["normalize"] is False

        # Check if the output is correct
        pd.testing.assert_frame_equal(output, expected_output)

    def test_aggregate_summary_df(global_summary_df, global_aggregate_summary_df):
        df = global_summary_df
        expected_output = global_aggregate_summary_df
        pd.set_option("display.max_columns", None)

        pd.testing.assert_frame_equal(InspectorShap.aggregate_summary_df(df), expected_output)

    def test_compute__report_done():
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        report_value = pd.DataFrame([[1, 2], [2, 3]], columns=["cluster_id", "column_a"])
        inspector.cluster_report = report_value

        pd.testing.assert_frame_equal(inspector.compute(), report_value)

    def test_compute__single_df(global_mock_summary_df):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        inspector.hasmultiple_dfs = False

        report_value = global_mock_summary_df

        def mock_compute_report(self):
            self.agg_summary_df = report_value

        with patch.object(InspectorShap, "_compute_report", mock_compute_report):
            output = inspector.compute()

        # Check output and side effects
        pd.testing.assert_frame_equal(output, report_value)
        pd.testing.assert_frame_equal(inspector.cluster_report, report_value)
        pd.testing.assert_frame_equal(inspector.agg_summary_df, report_value)

    def test_compute__multiple_df(global_mock_summary_df, global_mock_aggregate_summary_dfs):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        inspector.hasmultiple_dfs = True

        report_value = global_mock_summary_df
        inspector.agg_summary_dfs = global_mock_aggregate_summary_dfs

        expected_result = pd.DataFrame(
            [[1, 2, 3, 2], [2, 3, 3, 3]], columns=["cluster_id", "column_a", "column_a_sample_1", "column_a_sample_2"]
        )

        def mock_compute_report(self):
            self.agg_summary_df = report_value

        with patch.object(InspectorShap, "_compute_report", mock_compute_report):
            output = inspector.compute()

        # Check output and side effects
        pd.testing.assert_frame_equal(output, expected_result)
        pd.testing.assert_frame_equal(inspector.cluster_report, expected_result)
        pd.testing.assert_frame_equal(inspector.agg_summary_df, report_value)

    def test_compute__multiple_df_set_names(global_mock_summary_df, global_mock_aggregate_summary_dfs):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans", cluster_probability=False)
        inspector.hasmultiple_dfs = True
        inspector.set_names = ["suf1", "suf2"]

        report_value = global_mock_summary_df
        inspector.agg_summary_dfs = global_mock_aggregate_summary_dfs

        expected_result = pd.DataFrame(
            [[1, 2, 3, 2], [2, 3, 3, 3]], columns=["cluster_id", "column_a", "column_a_suf1", "column_a_suf2"]
        )

        def mock_compute_report(self):
            self.agg_summary_df = report_value

        with patch.object(InspectorShap, "_compute_report", mock_compute_report):
            output = inspector.compute()

        # Check output and side effects
        pd.testing.assert_frame_equal(output, expected_result)
        pd.testing.assert_frame_equal(inspector.cluster_report, expected_result)
        pd.testing.assert_frame_equal(inspector.agg_summary_df, report_value)

    def test_slice_cluster_no_inputs_not_complementary(
        global_summary_df, global_X_shap, global_y, global_predicted_proba
    ):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        summary = global_summary_df
        inspector.summary_df = summary
        inspector.cluster_report = summary
        inspector.X_shap = X_shap = global_X_shap
        inspector.y = y = global_y
        inspector.predicted_proba = predicted_proba = global_predicted_proba

        target_cluster_id = 1
        correct_mask = returned_mask = [True, False, False, False, True, False, False, False]
        inspector.get_cluster_mask.return_value = correct_mask

        with patch.object(InspectorShap, "compute") as mocked_compute:
            with patch.object(InspectorShap, "get_cluster_mask") as mock_get_cluster_mask:
                mock_get_cluster_mask.return_value = returned_mask
                shap_out, y_out, pred_out = inspector.slice_cluster(target_cluster_id, complementary=False)

                # Ensure mocked_compute not called
                mocked_compute.accert_not_called()
                # Ensure mock_get_cluster_mask called with correct arguments
                mock_get_cluster_mask.assert_called_once()
                pd.testing.assert_frame_equal(mock_get_cluster_mask.call_args[0][0], summary)
                assert mock_get_cluster_mask.call_args[0][1] == target_cluster_id

        # Check outputs
        pd.testing.assert_frame_equal(shap_out, X_shap[correct_mask])
        pd.testing.assert_series_equal(y_out, y[correct_mask])
        pd.testing.assert_series_equal(pred_out, predicted_proba[correct_mask])

    def test_slice_cluster_inputs_complementary(global_summary_df, global_X_shap, global_y, global_predicted_proba):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        summary = global_summary_df
        X_shap = global_X_shap
        y = global_y
        predicted_proba = global_predicted_proba

        target_cluster_id = 1
        correct_mask = np.array([False, True, True, True, False, True, True, True])
        returned_mask = np.logical_not(correct_mask)
        inspector.get_cluster_mask.return_value = correct_mask

        assert inspector.cluster_report is None

        def mock_compute(self):
            self.cluster_report = summary

        with patch.object(InspectorShap, "compute", mock_compute):
            with patch.object(InspectorShap, "get_cluster_mask") as mock_get_cluster_mask:
                mock_get_cluster_mask.return_value = returned_mask
                shap_out, y_out, pred_out = inspector.slice_cluster(
                    target_cluster_id,
                    summary_df=summary,
                    X_shap=X_shap,
                    y=y,
                    predicted_proba=predicted_proba,
                    complementary=True,
                )
                # Ensure mocked_get_cluster_mask called with correct arguments
                mock_get_cluster_mask.assert_called_once()
                pd.testing.assert_frame_equal(mock_get_cluster_mask.call_args[0][0], summary)
                assert mock_get_cluster_mask.call_args[0][1] == target_cluster_id

        # Check outputs and side effects
        pd.testing.assert_frame_equal(shap_out, X_shap[correct_mask])
        pd.testing.assert_series_equal(y_out, y[correct_mask])
        pd.testing.assert_series_equal(pred_out, predicted_proba[correct_mask])
        pd.testing.assert_frame_equal(inspector.cluster_report, summary)

    def test_init_inspector():
        mock_model = MockModel()
        inspector = InspectorShap(
            model=mock_model,
            algotype="kmeans",
            confusion_metric="proba",
            normalize_probability=True,
            cluster_probability=True,
        )
        assert inspector.model is mock_model
        assert inspector.isinspected is False
        assert inspector.hasmultiple_dfs is False
        assert inspector.normalize_proba is True
        assert inspector.cluster_probabilities is True
        assert inspector.agg_summary_df is None
        assert inspector.set_names is None
        assert inspector.confusion_metric == "proba"
        assert inspector.cluster_report is None
        assert inspector.y is None
        assert inspector.predicted_proba is None
        assert inspector.X_shap is None
        assert inspector.clusters is None
        assert inspector.algotype == "kmeans"
        assert inspector.fitted is False
        assert inspector.X_shaps == list()
        assert inspector.clusters_list == list()
        assert inspector.ys == list()
        assert inspector.predicted_probas == list()

    def test_init_inspector_error():
        with pytest.raises(NotImplementedError):
            InspectorShap(model=MockModel(), algotype="kmeans", confusion_metric="error")

    def test_init_inspector_error2():
        with pytest.raises(UnsupportedModelError):
            InspectorShap(model=MockModel(), algotype="error", confusion_metric="proba")

    def test_slice_cluster_eval_sets__single_df():
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        inspector.hasmultiple_dfs = False
        cluster_id = 1
        with pytest.raises(NotFittedError):
            inspector.slice_cluster_eval_set(cluster_id)

    def test_slice_cluster_eval_sets__multiple_df(
        global_X_shaps, global_ys, global_predicted_probas, global_summary_dfs
    ):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        inspector.hasmultiple_dfs = True

        inspector.X_shaps = X_shaps = global_X_shaps
        inspector.ys = ys = global_ys
        inspector.predicted_probas = predicted_probas = global_predicted_probas
        inspector.summary_dfs = summary_dfs = global_summary_dfs

        target_row = [0]
        target_cluster_id = 1
        target_complementary = False

        target_output = [
            [pd.DataFrame([[0, 3, 0]], columns=["shap_1", "shap_2", "shap_3"]), pd.Series([0]), pd.Series([0.1])],
            [pd.DataFrame([[0, 2, 0]], columns=["shap_1", "shap_2", "shap_3"]), pd.Series([1]), pd.Series([0.4])],
        ]

        with patch.object(InspectorShap, "slice_cluster") as mock_slice_cluster:
            # Setting multiple outputs
            mock_slice_cluster.side_effect = [
                (X_shaps[0].iloc[target_row], ys[0].iloc[target_row], predicted_probas[0].iloc[target_row]),
                (X_shaps[1].iloc[target_row], ys[1].iloc[target_row], predicted_probas[1].iloc[target_row]),
            ]

            output = inspector.slice_cluster_eval_set(target_cluster_id, complementary=target_complementary)

            # Check if inputs are correct at each call
            for call_index, call in enumerate(mock_slice_cluster.call_args_list):
                # On the position 1 of call there are kwargs
                assert call[1]["cluster_id"] == target_cluster_id
                assert call[1]["complementary"] == target_complementary
                pd.testing.assert_frame_equal(call[1]["summary_df"], summary_dfs[call_index])
                pd.testing.assert_frame_equal(call[1]["X_shap"], X_shaps[call_index])
                pd.testing.assert_series_equal(call[1]["predicted_proba"], predicted_probas[call_index])
                pd.testing.assert_series_equal(call[1]["y"], ys[call_index])

        # Check lengths of lists
        assert len(output) is len(target_output)

        # Go over the output and check each element
        for index, current_output in enumerate(output):
            pd.testing.assert_frame_equal(target_output[index][0], current_output[0])
            pd.testing.assert_series_equal(target_output[index][1], current_output[1])
            pd.testing.assert_series_equal(target_output[index][2], current_output[2])

    def test_compute_report_single_df(
        global_clusters, global_y, global_predicted_proba, global_summary_df, global_aggregate_summary_df
    ):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        inspector.hasmultiple_dfs = False
        inspector.normalize_proba = target_normalize = False

        inspector.clusters = input_clust = global_clusters
        inspector.y = input_y = global_y
        inspector.predicted_proba = input_predicted_proba = global_predicted_proba
        target_summary_df = global_summary_df
        aggregated_summary = global_aggregate_summary_df

        with patch.object(InspectorShap, "create_summary_df") as mock_create_summary_df:
            with patch.object(InspectorShap, "aggregate_summary_df") as mock_aggregate_summary_df:
                mock_create_summary_df.return_value = target_summary_df
                mock_aggregate_summary_df.return_value = aggregated_summary

                inspector._compute_report()

                # check if the methods were called with correct arguments
                pd.testing.assert_frame_equal(mock_aggregate_summary_df.call_args[0][0], target_summary_df)
                pd.testing.assert_series_equal(mock_create_summary_df.call_args[0][0], input_clust)
                pd.testing.assert_series_equal(mock_create_summary_df.call_args[0][1], input_y)
                pd.testing.assert_series_equal(mock_create_summary_df.call_args[0][2], input_predicted_proba)
                assert mock_create_summary_df.call_args[1]["normalize"] == target_normalize

        # Check if the function correctly stored variables
        pd.testing.assert_frame_equal(inspector.agg_summary_df, aggregated_summary)
        pd.testing.assert_frame_equal(inspector.summary_df, target_summary_df)

    def test_compute_report_multiple_df(
        global_clusters,
        global_y,
        global_predicted_proba,
        global_summary_df,
        global_aggregate_summary_df,
        global_summary_dfs,
        global_ys,
        global_predicted_probas,
        global_aggregate_summary_dfs_eval_set,
    ):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        inspector.hasmultiple_dfs = True
        inspector.normalize_proba = False

        inspector.clusters = global_clusters
        inspector.y = global_y
        inspector.predicted_proba = global_predicted_proba
        inspector.ys = global_ys
        inspector.predicted_probas = global_predicted_probas
        target_summary_df = global_summary_df
        target_summary_dfs = global_summary_dfs
        aggregated_summary_df = global_aggregate_summary_df
        aggregated_summary_dfs = global_aggregate_summary_dfs_eval_set

        with patch.object(InspectorShap, "create_summary_df") as mock_create_summary_df:
            with patch.object(InspectorShap, "aggregate_summary_df") as mock_aggregate_summary_df:
                # Set returns for each call of methods
                mock_create_summary_df.side_effect = [target_summary_df, target_summary_dfs[0], target_summary_dfs[1]]
                mock_aggregate_summary_df.side_effect = [
                    aggregated_summary_df,
                    aggregated_summary_dfs[0],
                    aggregated_summary_dfs[1],
                ]
                inspector._compute_report()

        assert inspector.agg_summary_df.equals(aggregated_summary_df)
        assert inspector.summary_df.equals(target_summary_df)
        for index, item in inspector.agg_summary_dfs:
            assert item.equals(aggregated_summary_dfs[index])
        for index, item in inspector.summary_dfs:
            assert item.equals(target_summary_dfs[index])

    def test_perform_fit_calc(global_X, global_y, global_predicted_proba, global_X_shap, global_clusters):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        inspector.model = MockModel()
        input_X = global_X
        input_y = global_y
        input_predicted_proba = global_predicted_proba
        values_probabilities = input_predicted_proba.tolist()

        def mock_fit_clusters(self, X_shap):
            inspector.fitted = True

        with patch.object(InspectorShap, "assert_is_dataframe") as mock_assert_is_dataframe:
            with patch.object(InspectorShap, "assert_is_series") as mock_assert_is_series:
                with patch.object(InspectorShap, "compute_probabilities") as mock_compute_probabilities:
                    with patch("probatus.interpret._shap_helpers.shap_to_df") as mock_shap_to_df:
                        with patch.object(InspectorShap, "fit_clusters", mock_fit_clusters):
                            with patch.object(InspectorShap, "predict_clusters") as mock_predict_clusters:
                                mock_assert_is_dataframe.return_value = input_X
                                mock_assert_is_series.return_value = input_y
                                mock_compute_probabilities.return_value = values_probabilities
                                mock_shap_to_df.return_value = global_X_shap
                                mock_predict_clusters.return_value = global_clusters.tolist()

                                out_y, out_predicted_proba, out_X_shap, out_clusters = inspector.perform_fit_calc(
                                    input_X, input_y, fit_clusters=True
                                )

        pd.testing.assert_series_equal(out_y, input_y)
        pd.testing.assert_series_equal(out_predicted_proba, input_predicted_proba)
        pd.testing.assert_frame_equal(out_X_shap, global_X_shap)
        pd.testing.assert_series_equal(out_clusters, global_clusters)
        assert inspector.fitted is True

    def test_fit__multiple_df(
        global_X,
        global_y,
        global_predicted_proba,
        global_X_shap,
        global_clusters,
        global_Xs,
        global_ys,
        global_predicted_probas,
        global_clusters_eval_set,
        global_X_shaps,
    ):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        input_eval_set = [(global_Xs[0], global_ys[0]), (global_Xs[1], global_ys[1])]
        input_sample_names = ["set1", "set2"]
        input_X = global_X
        input_y = global_y

        with patch.object(InspectorShap, "perform_fit_calc") as mock_perform_fit_calc:
            with patch.object(InspectorShap, "init_eval_set_report_variables") as mock_init_variables:
                mock_perform_fit_calc.side_effect = [
                    (global_y, global_predicted_proba, global_X_shap, global_clusters),
                    (global_ys[0], global_predicted_probas[0], global_X_shaps[0], global_clusters_eval_set[0]),
                    (global_ys[1], global_predicted_probas[1], global_X_shaps[1], global_clusters_eval_set[1]),
                ]

                inspector.fit(X=input_X, y=input_y, eval_set=input_eval_set, sample_names=input_sample_names)
                mock_init_variables.assert_called_once()

        assert inspector.hasmultiple_dfs is True
        assert inspector.set_names is input_sample_names
        assert inspector.y.equals(global_y)
        assert inspector.predicted_proba.equals(global_predicted_proba)
        assert inspector.X_shap.equals(global_X_shap)
        assert inspector.clusters.equals(global_clusters)
        assert all([a.equals(b) for a, b in zip(inspector.clusters_list, global_clusters_eval_set)])
        assert all([a.equals(b) for a, b in zip(inspector.X_shaps, global_X_shaps)])
        assert all([a.equals(b) for a, b in zip(inspector.predicted_probas, global_predicted_probas)])
        assert all([a.equals(b) for a, b in zip(inspector.ys, global_ys)])
        assert input_sample_names is inspector.set_names

    def test_compute_probabilities(global_X):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        input_X = global_X
        model_probas = np.array(
            [[0.2, 0.8], [0.7, 0.3], [0.7, 0.3], [0.3, 0.7], [0.2, 0.8], [0.3, 0.7], [0.7, 0.3], [0.5, 0.5]]
        )
        expected_output = np.array([0.8, 0.3, 0.3, 0.7, 0.8, 0.7, 0.3, 0.5])

        with patch.object(MockModel, "predict_proba") as mock_predict_proba:
            mock_predict_proba.return_value = model_probas
            np.testing.assert_array_equal(expected_output, inspector.compute_probabilities(input_X))

    def test_fit_compute(global_X, global_aggregate_summary_df):
        inspector = InspectorShap(model=MockModel(), algotype="kmeans")
        input_X = global_X
        expected_output = global_aggregate_summary_df

        with patch.object(InspectorShap, "fit") as mock_fit:
            with patch.object(InspectorShap, "compute") as mock_compute:
                mock_compute.return_value = global_aggregate_summary_df

                output = inspector.fit_compute(input_X)

                # Check if fit called with input X
                pd.testing.assert_frame_equal(mock_fit.call_args[0][0], input_X)
        # Check if the returned value correct
        pd.testing.assert_frame_equal(expected_output, output)



import os

import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import get_scorer
from sklearn.model_selection import RandomizedSearchCV, StratifiedGroupKFold, StratifiedKFold
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from probatus.feature_elimination import EarlyStoppingShapRFECV, ShapRFECV
from probatus.utils import preprocess_labels


@pytest.fixture(scope="function")
def X():
    """
    Fixture for X.
    """
    return pd.DataFrame(
        {
            "col_1": [1, 1, 1, 1, 1, 1, 1, 0],
            "col_2": [0, 0, 0, 0, 0, 0, 0, 1],
            "col_3": [1, 0, 1, 0, 1, 0, 1, 0],
        },
        index=[1, 2, 3, 4, 5, 6, 7, 8],
    )


@pytest.fixture(scope="session")
def catboost_classifier_class():
    """This fixture allows to reuse the import of the CatboostClassifier class across different tests.

    It is equivalent to importing the package at the beginning of the file.

    Importing catboost multiple times results in a ValueError: I/O operation on closed file.

    """
    from catboost import CatBoostClassifier

    return CatBoostClassifier


@pytest.fixture(scope="function")
def y():
    """
    Fixture for y.
    """
    return pd.Series([1, 0, 1, 0, 1, 0, 1, 0], index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope="function")
def sample_weight():
    """
    Fixture for sample_weight.
    """
    return pd.Series([1, 1, 1, 1, 1, 1, 1, 1], index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope="function")
def groups():
    """
    Fixture for groups.
    """
    return pd.Series(["grp1", "grp1", "grp1", "grp1", "grp2", "grp2", "grp2", "grp2"], index=[1, 2, 3, 4, 5, 6, 7, 8])


def test_shap_rfe_randomized_search(X, y, capsys):
    """
    Test with RandomizedSearchCV.
    """
    clf = DecisionTreeClassifier(max_depth=1)
    param_grid = {"criterion": ["gini"], "min_samples_split": [1, 2]}
    search = RandomizedSearchCV(clf, param_grid, cv=2, n_iter=2)
    shap_elimination = ShapRFECV(search, step=0.8, cv=2, scoring="roc_auc", n_jobs=4, random_state=1)
    report = shap_elimination.fit_compute(X, y)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    assert report.shape[0] == 2
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_rfe(X, y, sample_weight, capsys):
    """
    Test with ShapRFECV.
    """
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    shap_elimination = ShapRFECV(
        clf,
        random_state=1,
        step=1,
        cv=2,
        scoring="roc_auc",
        n_jobs=4,
    )
    shap_elimination = shap_elimination.fit(X, y, sample_weight=sample_weight, approximate=True, check_additivity=False)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_rfe_group_cv(X, y, groups, sample_weight, capsys):
    """
    Test ShapRFECV with StratifiedGroupKFold.
    """
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    cv = StratifiedGroupKFold(n_splits=2, shuffle=True, random_state=1)
    shap_elimination = ShapRFECV(
        clf,
        random_state=1,
        step=1,
        cv=cv,
        scoring="roc_auc",
        n_jobs=4,
    )
    shap_elimination = shap_elimination.fit(
        X, y, groups=groups, sample_weight=sample_weight, approximate=True, check_additivity=False
    )

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_pipeline_error(X, y, capsys):
    """
    Test with ShapRFECV for pipelines.
    """
    clf = Pipeline(
        [
            ("scaler", StandardScaler()),
            ("dt", DecisionTreeClassifier(max_depth=1, random_state=1)),
        ]
    )
    with pytest.raises(TypeError):
        shap_elimination = ShapRFECV(
            clf,
            random_state=1,
            step=1,
            cv=2,
            scoring="roc_auc",
            n_jobs=4,
        )
        shap_elimination = shap_elimination.fit(X, y, approximate=True, check_additivity=False)


def test_shap_rfe_linear_model(X, y, capsys):
    """
    Test ShapRFECV with linear model.
    """
    clf = LogisticRegression(C=1, random_state=1)
    shap_elimination = ShapRFECV(clf, random_state=1, step=1, cv=2, scoring="roc_auc", n_jobs=4)
    shap_elimination = shap_elimination.fit(X, y)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_rfe_svm(X, y, capsys):
    """
    Test with ShapRFECV with SVM.
    """
    clf = SVC(C=1, kernel="linear", probability=True)
    shap_elimination = ShapRFECV(clf, random_state=1, step=1, cv=2, scoring="roc_auc", n_jobs=4)
    shap_elimination = shap_elimination.fit(X, y)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_rfe_cols_to_keep(X, y, capsys):
    """
    Test for shap_rfe_cv with features to keep parameter.
    """
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    shap_elimination = ShapRFECV(
        clf,
        random_state=1,
        step=2,
        cv=2,
        scoring="roc_auc",
        n_jobs=4,
        min_features_to_select=1,
    )
    shap_elimination = shap_elimination.fit(X, y, columns_to_keep=["col_2", "col_3"])

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 2
    reduced_feature_set = set(shap_elimination.get_reduced_features_set(num_features=2))
    assert reduced_feature_set == {"col_2", "col_3"}

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_shap_rfe_randomized_search_cols_to_keep(X, y, capsys):
    """
    Test with ShapRFECV with column to keep param.
    """
    clf = DecisionTreeClassifier(max_depth=1)
    param_grid = {"criterion": ["gini"], "min_samples_split": [1, 2]}
    search = RandomizedSearchCV(clf, param_grid, cv=2, n_iter=2)
    shap_elimination = ShapRFECV(search, step=0.8, cv=2, scoring="roc_auc", n_jobs=4, random_state=1)
    report = shap_elimination.fit_compute(X, y, columns_to_keep=["col_2", "col_3"])

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    assert report.shape[0] == 2
    reduced_feature_set = set(shap_elimination.get_reduced_features_set(num_features=2))
    assert reduced_feature_set == {"col_2", "col_3"}

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


def test_calculate_number_of_features_to_remove():
    """
    Test with ShapRFECV with n features to remove.
    """
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=10, num_features_to_remove=3, min_num_features_to_keep=5
    )
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=8, num_features_to_remove=5, min_num_features_to_keep=5
    )
    assert 0 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=5, num_features_to_remove=1, min_num_features_to_keep=5
    )
    assert 4 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=5, num_features_to_remove=7, min_num_features_to_keep=1
    )


def test_get_feature_shap_values_per_fold(X, y):
    """
    Test with ShapRFECV with features per fold.
    """
    clf = DecisionTreeClassifier(max_depth=1)
    shap_elimination = ShapRFECV(clf)
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        clf,
        train_index=[2, 3, 4, 5, 6, 7],
        val_index=[0, 1],
        scorer=get_scorer("roc_auc"),
    )
    assert test_score == 1
    assert train_score > 0.9
    assert shap_values.shape == (2, 3)


def test_shap_rfe_same_features_are_kept_after_each_run():
    """
    Test a use case which appears to be flickering with Probatus 1.8.9 and lower.

    Expected result: every run the same outcome.
    Probatus <= 1.8.9: A different order every time.
    """
    SEED = 1234

    feature_names = [(f"f{num}") for num in range(1, 21)]

    # Code from tutorial on probatus documentation
    X, y = make_classification(
        n_samples=100,
        class_sep=0.05,
        n_informative=6,
        n_features=20,
        random_state=SEED,
        n_redundant=10,
        n_clusters_per_class=1,
    )
    X = pd.DataFrame(X, columns=feature_names)

    random_forest = RandomForestClassifier(
        random_state=SEED,
        n_estimators=70,
        max_features="log2",
        criterion="entropy",
        class_weight="balanced",
    )

    shap_elimination = ShapRFECV(
        clf=random_forest,
        step=0.2,
        cv=5,
        scoring="f1_macro",
        n_jobs=1,
        random_state=SEED,
    )

    report = shap_elimination.fit_compute(X, y, check_additivity=True, seed=SEED)
    # Return the set of features with the best validation accuracy

    kept_features = list(report.iloc[[report["val_metric_mean"].idxmax() - 1]]["features_set"].to_list()[0])

    # Results from the first run
    assert ["f6", "f10", "f12", "f14", "f15", "f17", "f18", "f20"] == kept_features


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_complex_dataset(complex_data, complex_lightgbm):
    """
    Test on complex dataset.
    """
    X, y = complex_data

    param_grid = {
        "n_estimators": [5, 7, 10],
        "num_leaves": [3, 5, 7, 10],
    }
    search = RandomizedSearchCV(complex_lightgbm, param_grid, n_iter=1)

    shap_elimination = ShapRFECV(clf=search, step=1, cv=10, scoring="roc_auc", n_jobs=3, verbose=50)

    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == X.shape[1]


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_shap_rfe_early_stopping_lightGBM(complex_data, capsys):
    """
    Test EarlyStoppingShapRFECV with a LGBMClassifier.
    """
    from lightgbm import LGBMClassifier

    clf = LGBMClassifier(n_estimators=200, max_depth=3)
    X, y = complex_data

    shap_elimination = EarlyStoppingShapRFECV(
        clf,
        random_state=1,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    shap_elimination = shap_elimination.fit(X, y, approximate=False, check_additivity=False)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_shap_rfe_early_stopping_XGBoost(complex_data, capsys):
    """
    Test EarlyStoppingShapRFECV with a LGBMClassifier.
    """
    from xgboost import XGBClassifier

    clf = XGBClassifier(n_estimators=200, max_depth=3, random_state=42)
    X, y = complex_data
    X["f1_categorical"] = X["f1_categorical"].astype(float)

    shap_elimination = EarlyStoppingShapRFECV(
        clf,
        random_state=1,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    shap_elimination = shap_elimination.fit(X, y, approximate=False, check_additivity=False)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1) == ["f4"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


# For now this test fails, catboost has issues with categorical variables and
@pytest.mark.xfail
@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_shap_rfe_early_stopping_CatBoost(complex_data, capsys, catboost_classifier_class):
    """
    Test EarlyStoppingShapRFECV with a CatBoostClassifier.
    """
    clf = catboost_classifier_class(random_seed=42)
    X, y = complex_data

    shap_elimination = EarlyStoppingShapRFECV(
        clf,
        random_state=1,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    shap_elimination = shap_elimination.fit(X, y, approximate=False, check_additivity=False)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    report = shap_elimination.compute()

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1)[0] in ["f4", "f5"]

    _ = shap_elimination.plot(show=False)

    # Check if there is any prints
    out, _ = capsys.readouterr()
    assert len(out) == 0


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_shap_rfe_randomized_search_early_stopping_lightGBM(complex_data):
    """
    Test EarlyStoppingShapRFECV with RandomizedSearchCV and a LGBMClassifier on complex dataset.
    """
    from lightgbm import LGBMClassifier

    clf = LGBMClassifier(n_estimators=200)
    X, y = complex_data

    param_grid = {
        "max_depth": [3, 4, 5],
    }
    search = RandomizedSearchCV(clf, param_grid, cv=2, n_iter=2)
    shap_elimination = EarlyStoppingShapRFECV(
        search,
        step=1,
        cv=10,
        scoring="roc_auc",
        early_stopping_rounds=5,
        eval_metric="auc",
        n_jobs=4,
        verbose=50,
        random_state=1,
    )
    report = shap_elimination.fit_compute(X, y)

    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    assert report.shape[0] == X.shape[1]
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]

    _ = shap_elimination.plot(show=False)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_get_feature_shap_values_per_fold_early_stopping_lightGBM(complex_data):
    """
    Test with ShapRFECV with features per fold.
    """
    from lightgbm import LGBMClassifier

    clf = LGBMClassifier(n_estimators=200, max_depth=3)
    X, y = complex_data
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(clf, early_stopping_rounds=5)
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        clf,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
        scorer=get_scorer("roc_auc"),
    )
    assert test_score > 0.6
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_get_feature_shap_values_per_fold_early_stopping_CatBoost(complex_data, catboost_classifier_class):
    """
    Test with ShapRFECV with features per fold.
    """
    clf = catboost_classifier_class(random_seed=42)
    X, y = complex_data
    X["f1_categorical"] = X["f1_categorical"].astype(str).astype("category")
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(clf, early_stopping_rounds=5)
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        clf,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
        scorer=get_scorer("roc_auc"),
    )
    assert test_score > 0
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_get_feature_shap_values_per_fold_early_stopping_XGBoost(complex_data):
    """
    Test with ShapRFECV with features per fold.
    """
    from xgboost import XGBClassifier

    clf = XGBClassifier(n_estimators=200, max_depth=3, random_state=42)
    X, y = complex_data
    X["f1_categorical"] = X["f1_categorical"].astype(float)
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(clf, early_stopping_rounds=5)
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        clf,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
        scorer=get_scorer("roc_auc"),
    )
    assert test_score > 0
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_EarlyStoppingShapRFECV_no_categorical(complex_data):
    """Test EarlyStoppingShapRFECV when no categorical features are present."""
    from lightgbm import LGBMClassifier

    model = LGBMClassifier(n_estimators=50, max_depth=3, num_leaves=3)

    shap_elimination = EarlyStoppingShapRFECV(
        clf=model,
        step=0.33,
        cv=5,
        scoring="accuracy",
        eval_metric="logloss",
        early_stopping_rounds=5,
    )
    X, y = complex_data
    X = X.drop(columns=["f1_categorical"])
    report = shap_elimination.fit_compute(X, y, feature_perturbation="tree_path_dependent")
    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    assert report.shape[0] == X.shape[1]
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]

    _ = shap_elimination.plot(show=False)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
def test_LightGBM_stratified_kfold():
    """
    Test added to check for https://github.com/ing-bank/probatus/issues/170.
    """
    from lightgbm import LGBMClassifier

    X = pd.DataFrame(
        [
            [1, 2, 3, 4, 5, 101, 102, 103, 104, 105],
            [-1, -2, 2, -5, -7, 1, 2, 5, -1, 3],
            ["a", "b"] * 5,  # noisy categorical will dropped first
        ]
    ).transpose()
    X[2] = X[2].astype("category")
    X[1] = X[1].astype("float")
    X[0] = X[0].astype("float")
    y = [0] * 5 + [1] * 5

    model = LGBMClassifier()
    n_iter = 2
    n_folds = 3

    for _ in range(n_iter):
        skf = StratifiedKFold(n_folds, shuffle=True, random_state=42)
        shap_elimination = EarlyStoppingShapRFECV(
            clf=model,
            step=1 / (n_iter + 1),
            cv=skf,
            scoring="accuracy",
            eval_metric="logloss",
            early_stopping_rounds=5,
        )
        report = shap_elimination.fit_compute(X, y, feature_perturbation="tree_path_dependent")
    assert shap_elimination.fitted
    shap_elimination._check_if_fitted()

    assert report.shape[0] == X.shape[1]

    shap_elimination.plot(show=False)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class NotFittedError(Exception):
    """
    Error.
    """

    def __init__(self, message):
        """
        Init error.
        """
        self.message = message


class DimensionalityError(Exception):
    """
    Error.
    """

    def __init__(self, message):
        """
        Init error.
        """
        self.message = message


class UnsupportedModelError(Exception):
    """
    Error.
    """

    def __init__(self, message):
        """
        Init error.
        """
        self.message = message


class NotInstalledError:
    """
    Raise error when a dependency is not installed.

    This object is used for optional dependencies.
    This allows us to give a friendly message to the user that they need to install extra dependencies as well as a link
    to our documentation page.

    Adapted from: https://github.com/RasaHQ/whatlies/blob/master/whatlies/error.py

    Example usage:

    ```python
    from probatus.utils import NotInstalledError
    try:
        import dash_core_components as dcc
    except ModuleNotFoundError as e:
        dcc = NotInstalledError("dash_core_components", "dashboard")
    dcc.Markdown() # Will raise friendly error with instructions how to solve
    ```
    Note that installing optional dependencies in a package are defined in setup.py.
    """

    def __init__(self, tool, dep=None):
        """
        Initialize error with missing package and reference to conditional install package.

        Args:
            tool (str): The name of the pypi package that is missing
            dep (str): The name of the extra_imports set (defined in setup.py) where the package is present. (optional)
        """
        self.tool = tool
        self.dep = dep

        msg = f"In order to use {self.tool} you'll need to install via;\n\n"
        if self.dep is None:
            msg += f"pip install {self.tool}\n\n"
        else:
            msg += f"pip install probatus[{self.dep}]\n\n"

        msg += "See probatus installation guide here: https://ing-bank.github.io/probatus/index.html"
        self.msg = msg

    def __getattr__(self, *args, **kwargs):
        """Raise when accessing an attribute."""
        raise ModuleNotFoundError(self.msg)

    def __call__(self, *args, **kwargs):
        """Raise when accessing a method."""
        raise ModuleNotFoundError(self.msg)



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numbers
import warnings

import numpy as np
import pandas as pd

from probatus.utils import DimensionalityError


def check_1d(x):
    """
    Checks whether or not a list, numpy array, pandas dataframe, pandas series are one-dimensional.

    Returns True when check is ok, otherwise throws a `DimensionalityError`

    Args:
        x: list, numpy array, pandas dataframe, pandas series

    Returns: True or throws `DimensionalityError`

    """
    if isinstance(x, list):
        if any([isinstance(el, list) for el in x]):
            raise DimensionalityError("The input is not 1D")
        else:
            return True
    if isinstance(x, np.ndarray):
        if x.ndim == 1 and all([isinstance(el, numbers.Number) for el in x]):
            return True
        else:
            raise DimensionalityError("The input is not 1D")
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1 and pd.api.types.is_numeric_dtype(x[x.columns[0]]):
            return True
        else:
            raise DimensionalityError("The input is not 1D")
    if isinstance(x, pd.core.series.Series):
        if x.ndim == 1 and pd.api.types.is_numeric_dtype(x):
            return True
        else:
            raise DimensionalityError("The input is not 1D")


def assure_numpy_array(x, assure_1d=False):
    """
    Returns x as numpy array. X can be a list, numpy array, pandas dataframe, pandas series.

    Args:
        x: list, numpy array, pandas dataframe, pandas series
        assure_1d: whether or not to assure that the input x is one-dimensional

    Returns: numpy array

    """
    if assure_1d:
        _ = check_1d(x)
    if isinstance(x, list):
        return np.array(x)
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            return x.values.flatten()
        else:
            return x.values
    if isinstance(x, pd.core.series.Series):
        return x.values


def assure_pandas_df(x, column_names=None):
    """
    Returns x as pandas DataFrame. X can be a list, list of lists, numpy array, pandas DataFrame or pandas Series.

    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested

    Returns:
        pandas DataFrame
    """
    if isinstance(x, pd.DataFrame):
        # Check if column_names are passed correctly
        if column_names is not None:
            x.columns = column_names
        return x
    elif any(
        [
            isinstance(x, np.ndarray),
            isinstance(x, pd.core.series.Series),
            isinstance(x, list),
        ]
    ):
        return pd.DataFrame(x, columns=column_names)
    else:
        raise TypeError("Please supply a list, numpy array, pandas Series or pandas DataFrame")


def assure_pandas_series(x, index=None):
    """
    Returns x as pandas Series. X can be a list, numpy array, or pandas Series.

    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested

    Returns:
        pandas Series
    """
    if isinstance(x, pd.Series):
        if isinstance(index, list) or isinstance(index, np.ndarray):
            index = pd.Index(index)
        current_x_index = pd.Index(x.index.values)
        if current_x_index.equals(index):
            # If exact match then keep it as it is
            return x
        elif current_x_index.sort_values().equals(index.sort_values()):
            # If both have the same values but in different order, then reorder
            return x[index]
        else:
            # If indexes have different values, overwrite
            x.index = index
            return x
    elif any([isinstance(x, np.ndarray), isinstance(x, list)]):
        return pd.Series(x, index=index)
    else:
        raise TypeError("Please supply a list, numpy array, pandas Series")


def check_numeric_dtypes(x):
    """
    Checks if all entries in an array are of a data type that can be interpreted as numeric (int, float or bool).

    Args:
        x (np.ndarray or pd.Series, list): array to be checked

    Returns:
        x: unchanged input array

    Raises:
        TypeError: if not all elements are of numeric dtypes
    """
    x = assure_numpy_array(x)
    allowed_types = [bool, int, float]

    for element in np.nditer(x):
        if type(element.item()) not in allowed_types:
            raise TypeError("Please supply an array with only floats, ints or booleans")
    return x


def preprocess_data(X, X_name=None, column_names=None, verbose=0):
    """
    Preprocess data.

    Does basic preprocessing of the data: Transforms to DataFrame, Warns which features have missing variables,
    and transforms object dtype features to category type, such that LightGBM handles them by default.

    Args:
        X (pd.DataFrame, list of lists, np.array):
            Provided dataset.

        X_name (str, optional):
            Name of the X variable, that will be printed in the warnings.

        column_names (list of str, optional):
            List of feature names of the provided samples. If provided it will be used to overwrite the existing
            feature names. If not provided the existing feature names are used or default feature names are
            generated.

        verbose (int, optional):
            Controls verbosity of the output:

            - 0 - neither prints nor warnings are shown
            - 1 - 50 - only most important warnings regarding data properties are shown (excluding SHAP warnings)
            - 51 - 100 - shows most important warnings, prints of the feature removal process
            - above 100 - presents all prints and all warnings (including SHAP warnings).

    Returns:
        (pd.DataFrame):
            Preprocessed dataset.
    """
    if X_name is None:
        X_name = "X"

    # Make sure that X is a pd.DataFrame with correct column names
    X = assure_pandas_df(X, column_names=column_names)

    # Warn if missing
    columns_with_missing = [column for column in X.columns if X[column].isnull().values.any()]
    if len(columns_with_missing) > 0:
        if verbose > 0:
            warnings.warn(
                f"The following variables in {X_name} contains missing values {columns_with_missing}. "
                f"Make sure to impute missing or apply a model that handles them automatically."
            )

    # Warn if categorical features and change to category
    indices_categorical_features = [
        column[0] for column in enumerate(X.dtypes) if column[1].name in ["category", "object"]
    ]
    categorical_features = list(X.columns[indices_categorical_features])

    # Set categorical features type to category
    if len(categorical_features) > 0:
        if verbose > 0:
            warnings.warn(
                f"The following variables in {X_name} contains categorical variables: "
                f"{categorical_features}. Make sure to use a model that handles them automatically or "
                f"encode them into numerical variables."
            )

        # Ensure category dtype, to enable models e.g. LighGBM, handle them automatically
        for categorical_feature in categorical_features:
            if X[categorical_feature].dtype.name == "object":
                X[categorical_feature] = X[categorical_feature].astype("category")
    return X, X.columns.tolist()


def preprocess_labels(y, y_name=None, index=None, verbose=0):
    """
    Does basic preparation of the labels. Turns them into Series, and wars in case the target is not binary.

    Args:
    y (pd.Series, list, np.array):
        Provided labels.

    y_name (str, optional):
        Name of the y variable, that will be printed in the warnings.

    index (list of int or pd.Index, optional):
            The index correct index that should be used for y. In case y is a list or np.array, the index is set when
            creating pd.Series. In case it is pd.Series already, if the indexes consist of the same values, the y is
            going to be ordered based on provided index, otherwise, the current index of y is overwritten by index
            argument.

    verbose (int, optional):
        Controls verbosity of the output:

        - 0 - neither prints nor warnings are shown
        - 1 - 50 - only most important warnings regarding data properties are shown (excluding SHAP warnings)
        - 51 - 100 - shows most important warnings, prints of the feature removal process
        - above 100 - presents all prints and all warnings (including SHAP warnings).

    Returns:
        (pd.Series):
            Labels in the form of pd.Series.
    """
    if y_name is None:
        y_name = "y"

    # Make sure that y is a series with correct index
    y = assure_pandas_series(y, index=index)

    # Warn if not binary labels
    if len(y.unique()) != 2:
        if verbose > 0:
            warnings.warn(
                f"The labels in {y_name} contain {y.unique()} unique values. The features in probatus support"
                f" binary classification models, thus, the feature might not work correctly."
            )
    return y
